#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2023-01-17 21:55:23
# @Author  : HeLiang (helianghit@foxmail.com)
# @Link    : https://github.com/HeLiangHIT

import os
from ui.qt_main import main

if __name__ == '__main__':
    main()
